(* A Translation to Free Pascal on Mac OS X by The Creative CAT, 2007

 * copyright (c) 2001 Fabrice Bellard
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

/**
 * @file apiexample.c
 * avcodec API use example.
 *
 * Note that this library only handles codecs (mpeg, mpeg4, etc...),
 * not file formats (avi, vob, etc...). See library 'libavformat' for the
 * format handling
 *)

(*
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define PI 3.14159265358979323846

#ifdef HAVE_AV_CONFIG_H
#undef HAVE_AV_CONFIG_H
#endif

#include "avcodec.h"

#define INBUF_SIZE 4096
*)

{$MODE DELPHI}
{$LINKLIB libavcodec}
{$LINKLIB libavutil}

uses avutil, avcodec;

const
(*  #define PI 3.14159265358979323846 is not required
	because FPC RTL has function pi *)
  INBUF_SIZE = 4096;
  InputBufferSize = INBUF_SIZE + FF_INPUT_BUFFER_PADDING_SIZE;
type
  bytebuf = array[0..1] of byte;
  pbytebuf = ^bytebuf;
var
  filename : string;

{$I+} (* Turn on I/O Checking *)
(*
 * Audio encoding example
 *)
procedure audio_encode_example(filename : string);
var
  codec : PAVCodec;
  c : PAVCodecContext = nil;
  frame_size, i, j, dummy, inbuf_size, out_size, outbuf_size : integer;
  f : file;
  samples : array of word (* PWord *);
  t, tincr : double;
  outbuf : PByte;
begin
  writeln('Audio encoding');

  (* find the MP2 encoder *)
  codec := avcodec_find_encoder(CODEC_ID_MP2);
  if codec = nil then begin
    writeln(stderr, 'codec not found');
    halt(1)
  end;

  c := avcodec_alloc_context;

  (* put sample parameters *)
  with c^ do begin
    bit_rate := 64000;
    sample_rate := 44100;
    channels := 2
  end;

  (* open it *)
  if avcodec_open(c, codec) < 0 then begin
    writeln(stderr, 'could not open codec');
    halt(1)
  end;

  (* the codec gives us the frame size, in samples *)
  frame_size := c^.frame_size;
  inbuf_size := frame_size * 2 * c^.channels;
  SetLength(samples, inbuf_size);  (* using FPC dynamic array *)
  outbuf_size := 10000;
  GetMem(outbuf, outbuf_size);

  assign(f, filename);
  rewrite(f, 1);

(* in default, FPC run time would halt in I/O error. So we can safely omit
  following code:

  if (!f) {
        fprintf(stderr, "could not open %s\n", filename);
        exit(1);
    }
*)

  (* encode a single tone sound *)
  t := 0;
  tincr := 2 * PI * 440.0 / c^.sample_rate;  (* sinusoidal wave *)
  for i := 0 to 199 do begin
    for j := 0 to pred(frame_size) do begin
      samples[j*2] := round(sin(t) * 10000);
      samples[j*2+1] := samples[j*2];
      t := t + tincr
    end;
    (* encode the samples *)
    out_size := avcodec_encode_audio(c, outbuf, outbuf_size, @samples[0]);
				(* @samples <> @samples[0], for it's a dynamic array *)
    blockwrite(f, outbuf^, out_size, dummy)
  end;
  close(f);
  FreeMem(outbuf, outbuf_size);

  avcodec_close(c);
  av_free(c)				(* avutil *)
end;

(*
 * Audio decoding.
 *)
procedure audio_decode_example(outfilename, filename : string);
var
  codec : PAVCodec;
  c : PAVCodecContext = nil;
  out_size, size, len, dummy : integer;
  f, outfile : file;
  outbuf : PByte;
  inbuf : array[1..InputBufferSize] of byte;
  inbuf_ptr : cardinal;
begin
  writeln('Audio decoding');

  (* find the mpeg audio decoder *)
  codec := avcodec_find_decoder(CODEC_ID_MP2);
  if codec = nil then begin
    writeln(stderr, 'codec not found');
    halt(1)
  end;

  c := avcodec_alloc_context;

  (* open it *)
  if avcodec_open(c, codec) < 0 then begin
    writeln(stderr, 'could not open codec');
    halt(1)
  end;
  
  GetMem(outbuf, AVCODEC_MAX_AUDIO_FRAME_SIZE);

  assign(f, filename);
  reset(f, 1);
(*
    if (!f) {
        fprintf(stderr, "could not open %s\n", filename);
        exit(1);
    }
*)
  assign(outfile, outfilename);
  rewrite(outfile, 1);
(*
  if (!outfile) {
        av_free(c);
        exit(1);
    }
*)

  (* decode until eof *)
  fillchar(inbuf, sizeof(inbuf), 0);
  blockread(f, inbuf, INBUF_SIZE, size);
  while size <> 0 do begin
    inbuf_ptr := 1;
    repeat
      len := avcodec_decode_audio(c, PWord(outbuf), out_size,
                                       @inbuf[inbuf_ptr], size);
      if len < 0 then begin
        writeln(stderr, 'Error while decoding');
        halt(1)
      end;
      if out_size > 0 then
        (* if a frame has been decoded, output it *)
        blockwrite(outfile, outbuf^, out_size, dummy);
      dec(size, len);
	  inc(inbuf_ptr, len)
    until size < len;
    seek(f, filepos(f) - size);  (* re-read incomplete frame *)
    blockread(f, inbuf, INBUF_SIZE, size)
  end;

  close(outfile);
  close(f);
  FreeMem(outbuf, AVCODEC_MAX_AUDIO_FRAME_SIZE);

  avcodec_close(c);
  av_free(c)
end;

(*
 * Video encoding example
 *)
procedure video_encode_example(filename : string);
var
  codec : PAVCodec;
  c : PAVCodecContext = nil;
  i, out_size, size, x, y, outbuf_size, picturebuf_size, dummy : integer;
  f : file;
  picture : PAVFrame;
  outbuf, picture_buf : PByte;
  pbuf : array[0..3] of pbytebuf;
  endsequence : array[1..4] of byte = ($00, $00, $01, $B7);
begin
  writeln('Video encoding');

  (* find the mpeg1 video encoder *)
  codec := avcodec_find_encoder(CODEC_ID_MPEG1VIDEO);
  if codec = nil then begin
    writeln(stderr, 'could not open codec');
    halt(1)
  end;

  c := avcodec_alloc_context();
  picture := avcodec_alloc_frame();

  (* put sample parameters *)
  with c^ do begin
    bit_rate := 400000;
  (* resolution must be a multiple of two *)
    width := 352;
    height := 288;
  (* frames per second *)
    time_base.num := 1; (* (AVRational) {1, 25} *)
    time_base.den := 25;
    gop_size := 10; (* emit one intra frame every ten frames *)
    max_b_frames := 1;
    pix_fmt := PIX_FMT_YUV420P
  end;

  (* open it *)
  if avcodec_open(c, codec) < 0 then begin
    writeln(stderr, 'could not open codec');
    halt(1)
  end;
  
  assign(f, filename);
  rewrite(f, 1);
(*
    if (!f) {
        fprintf(stderr, "could not open %s\n", filename);
        exit(1);
    }
*)

  (* alloc image and output buffer *)
  outbuf_size := 100000;
  GetMem(outbuf, outbuf_size);
  size := c^.width * c^.height;
  picturebuf_size := size * 3 div 2;
       					(* size for YUV 420,
				   		   we can use 'div' here, for size is a multiple of 4 *)
  GetMem(picture_buf, picturebuf_size);

  with picture^ do begin
    data[0] := picture_buf;
    data[1] := data[0] + size;  (* FPC allows pointer arithmetics *)
    data[2] := data[1] + size div 4;  (* FPC allows pointer arithmetics *)
    linesize[0] := c^.width;
    linesize[1] := c^.width div 2;
    linesize[2] := c^.width div 2
  end;

  with picture^ do begin (* this is a Pascalish way to handle C pointers as arrays *)
    pbuf[0] := data[0];
    pbuf[1] := data[1];
    pbuf[2] := data[2]
  end;

  (* encode 1 second of video *)
  for i := 0 to 24 do begin
    flush(stdout);
  (* prepare a dummy image *)
  (* Y *)
    for y := 0 to pred(c^.height) do
      for x := 0 to pred(c^.width) do
        pbuf[0]^[y * picture^.linesize[0] + x] := x + y + i * 3;

  (* Cb and Cr *)
    for y := 0 to pred(c^.height div 2) do
      for x := 0 to pred(c^.width div 2) do begin
        pbuf[1]^[y * picture^.linesize[1] + x] := 128 + y + i * 2;
        pbuf[2]^[y * picture^.linesize[2] + x] := 64 + x + i * 5
	  end;

  (* encode the image *)
    out_size := avcodec_encode_video(c, outbuf, outbuf_size, picture);
    writeln('encoding frame ', i:3, ' (size= ', out_size:5, ' )');
    blockwrite(f, outbuf^, out_size, dummy)
  end;

  (* get the delayed frames *)
  i := 25;
  while out_size <> 0 do begin
    flush(stdout);

    out_size := avcodec_encode_video(c, outbuf, outbuf_size, nil);
    writeln('write frame ', i:3, ' (size= ', out_size:5, ' )');
    blockwrite(f, outbuf^, out_size, dummy);
    inc(i)
  end;

  (* add sequence end code to have a real mpeg file *)
  
  blockwrite(f, endsequence, 4, dummy);
  close(f);
  FreeMem(picture_buf, picturebuf_size);
  FreeMem(outbuf, outbuf_size);

  avcodec_close(c);
  av_free(c);
  av_free(picture);
  writeln
end;

(*
 * Video decoding example
 *)

procedure video_decode_example(outfilename, filename : string);
var
  codec : PAVCodec;
  c : PAVCodecContext = nil;
  frame, size, got_picture, len, dummy : integer;
  f : file;
  picture : PAVFrame;
  inbuf : array[1..InputBufferSize] of byte;
  inbuf_ptr : cardinal;
  buf : string; (* original length is 1024. if you need more length, use AnsiString *)

  procedure pgm_save(buf : pointer; wrap, xsize, ysize : integer; filename : string);
  var
    f : file of byte;
    t : text;
    i : integer;
    pbuf : pbytebuf;
  begin
    assign(t, filename);   (* open as text for convenience ;-) *)
    rewrite(t);
    writeln(t, 'P5');
    writeln(t, xsize, ' ', ysize);
    writeln(t, 255);
    close(t);
    assign(f, filename);   (* re-open as binary file *)
    reset(f, 1);
    seek(f, filesize(f));  (* seek to eof *)
    pbuf := buf;
    for i := 0 to pred(ysize) do  (* write buffer contents *)
      blockwrite(f, pbuf^[i * wrap], xsize, dummy);
    close(f)
  end;

  function fnstr(i : integer) : string;
  var s : string;
  begin
    str(i, s);
    fnstr := s
  end;

  function ReplaceWithNum(s : string; n : integer) : string;
  var i : integer;
  begin
    i := pos('%d', s);
    delete(s, i, 2);
    insert(fnstr(n), s, i);
    ReplaceWithNum := s
  end;

begin
  (* set end of buffer to 0 (this ensures that no overreading happens for damaged mpeg streams) *)
  fillchar(inbuf[succ(INBUF_SIZE)], FF_INPUT_BUFFER_PADDING_SIZE, 0);

  writeln('Video decoding');

  (* find the mpeg1 video decoder *)
  codec := avcodec_find_decoder(CODEC_ID_MPEG1VIDEO);
  if codec = nil then begin
    writeln(stderr, 'codec not found');
    halt(1)
  end;

  c := avcodec_alloc_context();
  picture := avcodec_alloc_frame();

  if (codec^.capabilities and CODEC_CAP_TRUNCATED) <> 0 then
    c^.flags := c^.flags or CODEC_FLAG_TRUNCATED; (* we do not send complete frames *)

  (* For some codecs, such as msmpeg4 and mpeg4, width and height
       MUST be initialized there because this information is not
       available in the bitstream. *)

  (* open it *)
  if avcodec_open(c, codec) < 0 then begin
    writeln(stderr, 'could not open codec');
    halt(1)
  end;

  (* the codec gives us the frame size, in samples *)

  assign(f, filename);
  reset(f, 1);
(*
    if (!f) {
        fprintf(stderr, "could not open %s\n", filename);
        exit(1);
    }
*)

  frame := 0;
  blockread(f, inbuf, INBUF_SIZE, size);
  while size <> 0 do begin

        (* NOTE1: some codecs are stream based (mpegvideo, mpegaudio)
           and this is the only method to use them because you cannot
           know the compressed data size before analysing it.

           BUT some other codecs (msmpeg4, mpeg4) are inherently frame
           based, so you must call them with all the data for one
           frame exactly. You must also initialize 'width' and
           'height' before initializing them. *)

        (* NOTE2: some codecs allow the raw parameters (frame size,
           sample rate) to be changed at any frame. We handle this, so
           you should also take care of it *)

        (* here, we use a stream based decoder (mpeg1video), so we
           feed decoder and see if it could decode a frame *)

    inbuf_ptr := 1;
    repeat
      len := avcodec_decode_video(c, picture, got_picture,
									@inbuf[inbuf_ptr], size);
      if len < 0 then begin 
        writeln(stderr, 'Error while decoding frame ', frame);
        halt(1)
      end;

      if got_picture <> 0 then begin
        writeln('saving frame ', frame : 3);
        flush(stdout);

  (* the picture is allocated by the decoder. no need to free it *)
        buf := ReplaceWithNum(outfilename, frame);
        pgm_save(picture^.data[0], picture^.linesize[0],
                           c^.width, c^.height, buf);
        inc(frame)
      end;
      dec(size, len);
      inc(inbuf_ptr, len)
    until size < len;
    seek(f, filepos(f) - size);  (* re-read incomplete frame *)
    blockread(f, inbuf, INBUF_SIZE, size)
  end;

    (* some codecs, such as MPEG, transmit the I and P frame with a
       latency of one frame. You must do the following to have a
       chance to get the last frame of the video *)
  len := avcodec_decode_video(c, picture, got_picture, nil, 0);
  if got_picture <> 0 then begin
    writeln('saving last frame ', frame);
    flush(stdout);

  (* the picture is allocated by the decoder. no need to free it *)
    buf :=  ReplaceWithNum(outfilename, frame);
    pgm_save(picture^.data[0], picture^.linesize[0], c^.width, c^.height, buf);
    inc(frame)
  end;
  close(f);

  avcodec_close(c);
  av_free(c);
  av_free(picture);
  writeln
end;

begin
  (* must be called before using avcodec lib *)
  avcodec_init();

  (* register all the codecs *)
  avcodec_register_all();

  if paramcount <= 1 then begin
    audio_encode_example('/tmp/test.mp2');
    audio_decode_example('/tmp/test.sw', '/tmp/test.mp2');

    video_encode_example('/tmp/test.mpg');
    filename := '/tmp/test.mpg'
  end else
    filename := paramstr(1);

    //    audio_decode_example("/tmp/test.sw", filename);
  video_decode_example('/tmp/test%d.pgm', filename)
end.
